from typing import Dict, List, Optional, Union, cast

from rich.console import Console, ConsoleOptions, RenderResult
from rich.tree import Tree as RichTree

from .protocol import TrioNursery, TrioTask

""" Generate Identical Fake data for developing & test
Provide a simpler interface for constructing trio task tree & specifying nodes by their name
"""


class FakeNode:
    """The underlying node for fake objects"""

    def __init__(self, name: str):
        self._children: List = []
        self._parent: Optional["FakeNode"] = None
        self._name: str = name

        # FakeTrioTask is able to reference node by it's name without using registry

        # Each node would register itself with to root node's dict, so we could referece
        # node by it's name
        self._tree_root: Optional["FakeNode"] = self
        self._nodes: Dict[str, "FakeNode"] = {}  # only used by the tree root node
        self.register(self, name=name)

    def __contains__(self, key: Union[str]) -> bool:
        if type(key) is str:
            return key in self._nodes
        else:
            NotImplementedError
        return False

    def get_task_node(self, node_name: str) -> "FakeTrioTask":
        return cast("FakeTrioTask", self._nodes[node_name])

    def get_nursery_node(self, node_name: str) -> "FakeTrioNursery":
        return cast("FakeTrioNursery", self._nodes[node_name])

    @property
    def is_root(self) -> bool:
        return self == self.tree_root

    def tree_add(self, node: "FakeNode", parent_name: str):
        parent: "FakeNode" = self.tree_root._nodes[parent_name]
        if node in parent._children:
            raise RuntimeError(f"Children alreay exists: {node._name}")
        parent._children.append(node)
        node._parent = parent
        # child would register itself to the root node while new root node is assigned
        node.tree_root = self.tree_root

    def tree_remove(self, node_name: str):
        node = self.tree_root._nodes[node_name]
        if node == self.tree_root:
            raise RuntimeError("Should not remove root from tree")
        # except the root node, every node must have parent
        if node._parent is None:
            raise RuntimeError("Bug: orphan node")
        parent = node._parent
        if node not in node._parent._children:
            raise RuntimeError("Bug: children node not recorded in parent")
        parent._children.remove(node)
        node._parent = None
        # Modify root
        node.tree_root = node

    @property
    def tree_root(self):
        return self._tree_root

    @tree_root.setter
    def tree_root(self, new_root: "FakeNode"):
        """ side effect: would register itself to the root node"""
        if new_root is None:
            raise RuntimeError("Should not assign root to none")
        if new_root == self._tree_root:
            return

        # parent management
        if new_root is self and self._parent is not None:
            parent = self._parent
            if self not in parent._children:
                raise RuntimeError("Bug: children node not recorded in parent")
            parent._children.remove(self)
            self._parent = None

        # only used for type checking...
        assert self._tree_root is not None

        # unregister from old root
        # clear self from the origianl tree root
        del self._tree_root._nodes[self._name]

        # register to the new root
        self._tree_root = new_root
        self._tree_root.register(node=self, name=self._name)

        # propagate to all of it's children
        for child in self._children:
            child.tree_root = self.tree_root

    def register(self, node: "FakeNode", name: str):
        """ Register a node to this root node"""
        if not self.is_root:
            raise RuntimeError("Could only call register on a root task")
        if name in self._nodes:
            raise ValueError(f"Registered name already exists: {name}")
        self._nodes[name] = node


class FakeTrioNursery(FakeNode, TrioNursery):
    """Represent the trio internal type."""

    def __init__(self, name: str):
        super().__init__(name)

    @property
    def child_tasks(self):
        return self._children

    @property
    def _trio_vis_name(self):
        return self._name

    def _add_task(self, task: "FakeTrioTask"):
        self.tree_root.tree_add(task, self._name)

    def _remove_task(self, task_name: str):
        self.tree_root.tree_remove(task_name)

    def __repr__(self):
        if len(self.child_tasks) > 0:
            return f"<Nursery({self._name}) {self.child_tasks}>"
        return f"<Nursery({self._name})>"

    @property
    def rich_tag(self):
        ret = "[bold magenta]Nursery"
        if self._name is not None:
            ret += f": [white]{self._name}"
        return ret

    def _rich_node(self, parent: RichTree):
        cur_node = parent.add(self.rich_tag)
        for t in self.child_tasks:
            t._rich_node(parent=cur_node)


class FakeTrioTask(FakeNode, TrioTask):
    """Represent the trio internal type
    The only usable attributes are `self.child_nurseries` & `name`
    """

    def __init__(self, name: str):
        super().__init__(name)
        print(f"get root-> {super().tree_root}")

    @property
    def name(self):
        return self._name

    @property
    def child_nurseries(self):
        return self._children

    def _add_nursery(self, nursery: "FakeTrioNursery"):
        self.tree_root.tree_add(nursery, parent_name=self._name)

    def _remove_nursery(self, nursery_name: str):
        self.tree_root.tree_remove(nursery_name)

    def __repr__(self):
        if len(self.child_nurseries) > 0:
            return f"<Task:{self.name} {self.child_nurseries}>"
        return f"<Task: {self.name}>"

    @property
    def rich_tag(self):
        ret = "[bold yellow]Task"
        if self.name is not None:
            ret += f": [white]{self.name}"
        return ret

    def _rich_node(self, parent: RichTree):
        cur_node = parent.add(self.rich_tag)
        [n._rich_node(parent=cur_node) for n in self.child_nurseries]

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        """Support Python `rich` lib"""
        root = RichTree(self.rich_tag)
        [n._rich_node(parent=root) for n in self.child_nurseries]
        yield root


def build_tree_from_json(dic: Dict) -> FakeTrioTask:
    """Generate a trio task tree based on dict provided"""

    def build_task(task: Dict) -> FakeTrioTask:
        t = FakeTrioTask(name=task.get("name", None))
        for n in task["nurseries"]:
            t._add_nursery(build_nursery(n))
        return t

    def build_nursery(nursery: Dict) -> FakeTrioNursery:
        n = FakeTrioNursery(name=nursery.get("name", None))
        for t in nursery["tasks"]:
            n._add_task(build_task(t))
        return n

    return build_task(dic)



from pathlib import Path
from typing import TYPE_CHECKING

from pydantic import BaseModel, validator

# only used for vscode intelliSense
# ref: https://github.com/microsoft/python-language-server/issues/1898
if TYPE_CHECKING:
    import dataclasses

    static_check_init_args = dataclasses.dataclass
else:

    def static_check_init_args(cls):
        return cls


@static_check_init_args
class VisConfig(BaseModel):

    # Ignore trio's internal nursery/task
    # the side effect is that we would inspect every task/nersery's code object
    # so we would like to disable this config while testing
    only_vis_user_scope: bool = True

    print_task_tree: bool = True

    log_overwrite_if_exists: bool = True
    log_filename: str = "./sc-logs.json"

    @validator("log_filename")
    def log_file_should_not_be_overwritten_unless_required(cls, filename: str, values):
        if Path(filename).exists() and values["log_overwrite_if_exists"] is False:
            raise ValueError(
                f"[trio-vis] file exists with overwrite set to false:{filename}"
            )



from typing import List

from typing_extensions import Protocol

""" Use Protocol for better type checking
"""


class TrioNursery(Protocol):
    """Represent trio's internal `Nursery` type"""

    child_tasks: List["TrioTask"]


class TrioTask(Protocol):
    """Represent trio's internal `Task` type"""

    name: str
    child_nurseries: List["TrioNursery"]


class TrioInstrument(Protocol):
    """Trio's instrument class to used the instrument API

    Refer to:
        https://trio.readthedocs.io/en/stable/reference-lowlevel.html#trio.abc.Instrument
    """

    def task_exited(self, task: TrioTask):
        pass

    def task_spawned(self, task: TrioTask):
        pass



from typing import Optional

import rich

from trio_vis.config import VisConfig
from trio_vis.desc_tree import DescNode, DescTree
from trio_vis.protocol import TrioInstrument, TrioNursery, TrioTask
from trio_vis.registry import SCRegistry
from trio_vis.sc_logger import Logger, SCLogger


def is_user_task(task):
    filename = None
    filename = task.coro.cr_frame.f_code.co_filename
    if "trio/_core" in filename:
        return False
    return True


"""Capture cases
+ -- task spawned
    1. Child root task spawned
    2. A new nursery started
    3. Task spawnd under a nursery
+ -- task exited
    4. Task exited but it's parent nursery still exists
    5. Task exited with it's child nursery ended
    6. root task exited
"""


def parent_from_registry(target, registry: SCRegistry):
    pass


class SC_Monitor(TrioInstrument):
    """SC Monitor
    Monitoring key structured-concurreny events happend in Trio
    """

    def __init__(self, config: VisConfig = VisConfig(), sc_logger=SCLogger):
        self.registry = SCRegistry()
        self.root_task: Optional[TrioTask] = None
        self.desc_tree: Optional[DescTree] = None
        self.root_exited = False

        self.cfg: VisConfig = config

        self.event_id: int = 0
        self.called_id: int = 0

        # the logger would write the entire file right before the program exit
        self.sc_logger: Logger = sc_logger(
            self.registry, log_filename=self.cfg.log_filename
        )

    def log(self, msg):
        self.event_id += 1
        if self.cfg.print_task_tree is True:
            rich.print(f"[red]event-{self.event_id}[reset] - [blue]{msg}")

    def api_called(self):
        if self.cfg.print_task_tree is True:
            rich.print(f"[yellow]api called-{self.called_id}[reset]")
        self.called_id += 1

    @classmethod
    def from_tree(cls, root_task: TrioTask, *args, **kwargs):
        """Build the Monitor with a given state, used for testing"""
        instance = cls(*args, **kwargs)
        instance.cfg.only_vis_user_scope = False
        instance.root_task = root_task
        instance.rebuild_tree()
        return instance

    def _name(self, obj) -> str:
        """Get parsed info from trio's task/nursery"""
        return self.registry.get_info(obj).name

    def rebuild_tree(self):
        self.desc_tree = DescTree.build(
            root_task=self.root_task, registry=self.registry
        )

    def task_spawned(self, task):
        if self.cfg.only_vis_user_scope is True and not is_user_task(task):
            return
        self.api_called()
        if not self.root_task:
            self.root_task = task
            self.desc_tree = DescTree.build(task, registry=self.registry)
            self.log(f"root task added:{self._name(task)}")
            self.sc_logger.log_start(child=task, parent=None)
            return

        # Rebuild: becuase parent nursery might be added without notify our monitor
        # in that case we might need to detect these changes ourself
        # However, to keep it simple, we simply rebuild the whole tree here
        self.rebuild_tree()
        if self.cfg.print_task_tree:
            rich.print(self.desc_tree)

        parent_nursery: TrioNursery = self.desc_tree.get_parent_ref(task)
        if parent_nursery is None:
            self.log("Error!!!")
            return

        if len(parent_nursery.child_tasks) == 1:
            grand_parent = self.desc_tree.get_parent_ref(parent_nursery)
            self.log(f"nursery started: {self._name(parent_nursery)}")
            self.sc_logger.log_start(child=parent_nursery, parent=grand_parent)

        self.log(
            f"task started: {self._name(task)}, parent:{self._name(parent_nursery)}"
        )
        self.sc_logger.log_start(child=task, parent=parent_nursery)

    def task_exited(self, task):
        # we only trace user task
        if self.root_exited:
            return

        self.api_called()
        task_name = self._name(task)

        desc_task: DescNode = self.desc_tree.ref_2node[task]
        if len(desc_task.children) > 0:
            for desc_child_nursery in desc_task.children:
                nursery_name = self._name(desc_child_nursery.ref)
                self.log(f"nursery exited: {nursery_name}, parent: {task_name}")
                self.sc_logger.log_exit(child=desc_child_nursery.ref, parent=task)
                self.desc_tree.remove_ref(desc_child_nursery.ref)

        if task == self.root_task:
            self.log(f"root task exited: {task_name}")
            self.root_exited = True
            self.sc_logger.log_exit(child=task, parent=None)
            # Notice: we shouldn't remove root task from registry
            # since we need to retrieve it's name from registry in the logger
            return

        parent_nursery: TrioNursery = self.desc_tree.get_parent_ref(task)
        self.log(f"task exited: {task_name}, parent:{self._name(parent_nursery)}")
        self.sc_logger.log_exit(child=task, parent=parent_nursery)
        self.desc_tree.remove_ref(task)

        self.rebuild_tree()
        if self.cfg.print_task_tree:
            rich.print(self.desc_tree)



# Logger for structure-concurrent events
import atexit
import json
from abc import abstractmethod
from typing import Dict, List, Optional, cast

import attr
from typing_extensions import Protocol

from .desc_tree import TrioNode
from .registry import SCRegistry


class Logger(Protocol):
    @abstractmethod
    def log_start(self, child: TrioNode, parent: Optional[TrioNode]):
        raise NotImplementedError

    @abstractmethod
    def log_exit(self, child: TrioNode, parent: Optional[TrioNode]):
        raise NotImplementedError


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SCEvent:
    time: int
    desc: str
    id: str
    parent: str

    def as_dict(self) -> Dict:
        return attr.asdict(self)


def get_name(target: Optional[TrioNode], registry: SCRegistry) -> str:
    if target == None:
        return "null"
    target = cast(TrioNode, target)
    return registry.get_info(target).name


class SCLogger(Logger):
    def __init__(self, registry: SCRegistry, log_filename: str):
        self.event_id: int = 0
        self.registry: SCRegistry = registry
        self.events: List[Dict] = []
        self.log_filename: str = log_filename
        atexit.register(self._write_log)

    def _get_id(self, child, parent):
        if child is None:
            raise RuntimeError("target should not be none")
        child_id = get_name(child, self.registry)
        parent_id = get_name(parent, self.registry)
        return (child_id, parent_id)

    def log_start(self, child: TrioNode, parent: Optional[TrioNode]):
        child_id, parent_id = self._get_id(child, parent)
        event = SCEvent(time=self.time, id=child_id, desc="created", parent=parent_id)
        self.events.append(event.as_dict())

    def log_exit(self, child: TrioNode, parent: Optional[TrioNode]):
        child_id, parent_id = self._get_id(child, parent)
        event = SCEvent(time=self.time, id=child_id, desc="exited", parent=parent_id)
        self.events.append(event.as_dict())

    def _write_log(self):
        with open(self.log_filename, "w") as log_file:
            json.dump(self.events, log_file, indent=4)

    @property
    def time(self) -> int:
        ret = self.event_id
        self.event_id += 1
        return ret


def log():
    event = SCEvent(2, "t0", "start", "n0")
    print(json.dumps(event.as_dict()))


def main():
    print("good")
    log()


if __name__ == "__main__":
    main()



from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, Optional, Union, cast

from .protocol import TrioNursery, TrioTask

"""
Name Trio objects
"""


class SerialNumberGen:
    """Give serial number to given names"""

    def __init__(self):
        self.numbers: Dict[str, int] = defaultdict(int)

    def draw(self, name: str) -> int:
        serial_num = self.numbers[name]
        self.numbers[name] += 1
        return serial_num


TYPE_TRIO_TASK = 1
TYPE_TRIO_NURSERY = 2
TYPE_UNKOWN = -1


def parse_obj_type(object):
    if getattr(object, "child_tasks", None) != None:
        return TYPE_TRIO_NURSERY
    elif getattr(object, "child_nurseries", None) != None:
        return TYPE_TRIO_TASK
    return TYPE_UNKOWN


# TODO: slotted? freezed?
@dataclass
class RegisteredSCInfo:
    """Store Information about a scope-like object (Task, Nursery)"""

    name: str
    serial_num: Optional[int]

    @classmethod
    def from_task(cls, task: TrioTask, serial_num: SerialNumberGen):
        num = serial_num.draw(task.name)
        name = f"{task.name}-{num}"
        return cls(name=name, serial_num=num)

    @classmethod
    def from_nursery(cls, _nursery: TrioNursery, serial_num: SerialNumberGen):
        # TODO: parse from more specific info
        num = serial_num.draw("nursery")

        self_defined_name = getattr(_nursery, "_trio_vis_name", None)
        if self_defined_name is not None:
            name = self_defined_name
        else:
            name = f"nursery-{num}"
        return cls(name=name, serial_num=num)


class SCRegistry:
    """Help to register Structured-Concurrency-related objects

    Also store some meaningful information about those objects
    inside the registry

    here we would register Trio's Task/Nursery object
    """

    def __init__(self):
        self.serial_drawer = SerialNumberGen()
        self.registered: Dict[Any, RegisteredSCInfo] = {}

    def get_name(self, obj: Union[TrioTask, TrioNursery]):
        if obj in self.registered:
            return self.registered[obj].name
        return self.get_info(obj).name

    def get_info(self, obj: Union[TrioTask, TrioNursery]) -> RegisteredSCInfo:
        """Get object information
        would register if not yet stored
        """
        if obj in self.registered:
            return self.registered[obj]

        obj_type = parse_obj_type(obj)
        info: RegisteredSCInfo
        if obj_type is TYPE_TRIO_TASK:
            info = RegisteredSCInfo.from_task(cast(TrioTask, obj), self.serial_drawer)
        elif obj_type is TYPE_TRIO_NURSERY:
            info = RegisteredSCInfo.from_nursery(
                cast(TrioNursery, obj), self.serial_drawer
            )
        else:
            raise RuntimeError("[registry] Unkown type")
        self.registered[obj] = info
        return info

    def remove(self, obj) -> bool:
        if obj in self.registered:
            del self.registered[obj]
            return True
        else:
            raise RuntimeError("Bug remove item not in registry")
